#ifndef MESH_H
#define MESH_H

#include <QtOpenGL>
#include <QOpenGLFunctions_4_3_Core>
#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <postprocess.h>
#include "texture.h"

#define positionIndex 0
#define normalIndex 1
#define uvIndex 2

#define EPSILON 0.00001

struct Vertex
{
    QVector3D pos;
    float padding0;

    QVector3D normal;
    float padding1;

    QVector2D tex;
    QVector2D padding2;


    Vertex(){}

    Vertex(const QVector3D& pos, const QVector3D& normal, const QVector2D& tex)
    {
        this->pos    = pos;
        this->tex    = tex;
        this->normal = normal;
    }

    bool samePos(const Vertex& other) const{
        return pos == other.pos;
    }

//TODO add some kind of toString so we can log vertices.
};

QDebug operator<< (QDebug d, const Vertex &v);

class Mesh
{
public:
    Mesh(QOpenGLFunctions_4_3_Core *f,QString filename);
    Mesh(QOpenGLFunctions_4_3_Core *f, Mesh *mesh, QVector<Vertex> &vertex_buffer, QVector<GLuint> &index_buffer);
    ~Mesh();

    void render(QOpenGLShaderProgram *shader, QMatrix4x4 V,QMatrix4x4 P, int subdivision);
    const aiScene *scene;
    bool debug = true;

    struct MeshEntry{
        MeshEntry();
        ~MeshEntry();
        QVector<unsigned int> getIndexBuffer();
        QVector<Vertex> getVertexBuffer();


        void init(QOpenGLFunctions_4_3_Core *f,QVector<Vertex>& Vertices,
                  QVector<unsigned int>& Indices);

        void update(GLuint VB_handle,QVector<Vertex>& Vertices, QVector<unsigned int>& Indices);

        QString name;
        GLuint VB_handle;
        GLuint IB_handle;

        QVector<unsigned int> indices;
        QVector<Vertex> vertices;

        int numIndex;
        int materialIndex;
        QOpenGLFunctions_4_3_Core *f;

        QVector3D min;
        QVector3D max;
    };

    struct MaterialInfo
    {
        MaterialInfo();
        QString Name;
        QVector3D Diffuse;
        QVector3D Specular;
        bool hasTexture;
        Texture texture;
        float Shininess;
    };

    struct Node
    {
        Node();
        QString name;
        QMatrix4x4 transformation;
        QVector<int> meshes;
        QVector<Node> children;

        bool getFirstMeshIndex(int& index);//false if no mesh entry found

    };


    Mesh::Node getRootNode();
    Mesh::MeshEntry *getMeshEntry(int index);

private:

    Assimp::Importer importer;
    QMatrix4x4 globalInverseTransform;
    QMatrix4x4 screenTransform;

    QVector<MeshEntry> entries;
    QVector<MaterialInfo> materials;

    Node rootNode;
    QOpenGLFunctions_4_3_Core *f;
    bool loaded;

    void initMeshEntry(int i,aiMesh * entry);
    void initMaterial(QString dir, unsigned int i, const aiMaterial* material);
    void initNode(const aiScene *scene, aiNode *node, Node &newNode, QString debugoffset);


    void renderNode(QOpenGLShaderProgram *shader, Node &node, QMatrix4x4 V,QMatrix4x4 P,QMatrix4x4 M, int subdivision);
    void renderMesh(QOpenGLShaderProgram *shader, int index, int subdivision);

    void findObjectDimension(Node node, QMatrix4x4 transform, QVector3D &min, QVector3D &max);


};

#endif // MESH_H
